/*
 * COPYRIGHT AND LICENSE
 * 
 * Copyright (c) 2004-2005, Juniper Networks, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *       1.      Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the following
 * disclaimer.
 *       2.      Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *       3.      The name of the copyright owner may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <event.h>

#include "amt.h"
#include "gw.h"

static const char __attribute__((unused)) id[] = "@(#) $Id: gw_sock.c,v 1.18 2005/05/05 18:35:33 pusateri Exp $";

int
socket_set_non_blocking(int s)
{
    int rc, val;

    val = fcntl(s, F_GETFL, 0);
    if (val < 0) {
	return errno;
    }
    rc = fcntl(s, F_SETFL, val | O_NONBLOCK);
    if (rc < 0) {
	return errno;
    }
    return 0;
}

static int
init_discovery_socket(gw_t *gw)
{
    int s, rc;
    struct sockaddr_in sin;

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0) {
	fprintf(stderr, "%s: creating discovery socket: %s\n", gw->name,
		strerror(errno));
	return errno;
    }
    gw->disco_sock = s;

    rc = socket_set_non_blocking(s);
    if (rc < 0) {
	return errno;
    }

    bzero(&sin, sizeof(sin));
    sin.sin_family = AF_INET;
#ifdef BSD	
    sin.sin_len = sizeof(sin);
#endif	
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
    sin.sin_port = htons(0);

    rc = bind(s, (struct sockaddr *) &sin, sizeof(sin));
    if (rc < 0) {
	fprintf(stderr, "%s: binding any on UDP socket: %s\n", gw->name,
		strerror(errno));
	return errno;
    }

    event_set(&gw->udp_disco_id, s, EV_READ|EV_PERSIST, gw_event_udp,
	      (void *) gw);
    rc = event_add(&gw->udp_disco_id, NULL);
    if (rc < 0) {
        fprintf(stderr, "%s: error from disco event_add: %s\n", gw->name,
		strerror(errno));
	return errno;
    }

    return 0;
}

static int
init_routing_socket(gw_t *gw)
{
    int s, rc;

    /*
     * initialize routing socket
     */
    s = socket(AF_ROUTE, SOCK_RAW, 0);
    if (s < 0) {
	fprintf(stderr, "%s: creating routing socket: %s\n", gw->name,
		strerror(errno));
	return errno;
    }
    rc = socket_set_non_blocking(s);
    if (rc < 0) {
	return errno;
    }
    gw->rt_sock = s;
    gw->seq = random();

    return 0;
}

/*
 * Create a UDP socket to communicate with the relay.
 * Determine our local address that we should use for sending packets
 * to the relay.
 * Add our new socket to the event queue.
 */
int
gw_init_udp_sock(gw_t *gw)
{
    int s, rc, sin_len;
    socklen_t len;
    struct sockaddr_in sin;

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0) {
	fprintf(stderr, "%s: creating UDP socket: %s\n", gw->name,
		strerror(errno));
	return errno;
    }
    gw->udp_sock = s;

    rc = socket_set_non_blocking(s);
    if (rc < 0) {
	return errno;
    }

#if 1
    bzero(&sin, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
    sin.sin_port = htons(0);

    rc = bind(s, (struct sockaddr *) &sin, sizeof(sin));
    if (rc < 0) {
	fprintf(stderr, "%s: binding any on UDP socket: %s\n", gw->name,
		strerror(errno));
	return errno;
    }
#endif

    bzero(&sin, sizeof(sin));
    sin.sin_family = AF_INET;
	sin_len = sizeof(sin);
#ifdef BSD	
    sin.sin_len = sizeof(sin);
#endif	
    sin.sin_addr.s_addr = gw->unicast_relay_addr.sin_addr.s_addr;
    sin.sin_port = htons(AMT_PORT);

    rc = connect(s, (struct sockaddr *) &sin, sizeof(sin));
    if (rc < 0) {
	/*
	 * we need to handle the case of no route to destination
	 * XXX
	 */
	if (errno == EADDRNOTAVAIL) {
	}
	fprintf(stderr, "%s: connecting on UDP socket: %s\n", gw->name,
		strerror(errno));
	return errno;
    }

    len = sizeof(sin);
    bzero(&sin, len);
    rc = getsockname(s, (struct sockaddr *) &sin, &len);
    if (rc < 0) {
	fprintf(stderr, "%s: getsockname: %s\n", gw->name,
		strerror(errno));
	return errno;
    }

    bcopy(&sin, &gw->local_addr, sin_len);

    event_set(&gw->udp_event_id, s, EV_READ|EV_PERSIST, gw_event_udp,
	      (void *) gw);
    rc = event_add(&gw->udp_event_id, NULL);
    if (rc < 0) {
        fprintf(stderr, "%s: error from udp event_add: %s\n", gw->name,
		strerror(errno));
	return errno;
    }

    return 0;
}

void
gw_cleanup_udp_sock(gw_t *gw)
{
    int rc;

    rc = event_del(&gw->udp_disco_id);
    if (rc < 0) {
        fprintf(stderr, "%s: error from disco event_del: %s\n", gw->name,
		strerror(errno));
    }
    bzero(&gw->udp_disco_id, sizeof(struct event));
    close(gw->disco_sock);
    gw->disco_sock = 0;
}

int
init_sockets(gw_t *gw)
{
    int rc;

    rc = init_discovery_socket(gw);
    if (rc) {
	return rc;
    }
    rc = init_routing_socket(gw);
    if (rc) {
	return rc;
    }
    return 0;
}
